@@ -24,6 +24,7 @@ These imports will not be delayed:
   b = __import__(a)
 '''
 
+import __builtin__
 _origimport = __import__
 
 class _demandmod(object):
@@ -116,6 +117,7 @@ ignore = [
     '_xmlplus',
     'fcntl',
     'win32com.gen_py',
+    'pythoncom',
     # imported by tarfile, not available under Windows
     'pwd',
     'grp',
@@ -126,9 +128,9 @@ ignore = [
 
 def enable():
     "enable global demand-loading of modules"
-    __builtins__["__import__"] = _demandimport
+    __builtin__.__import__ = _demandimport
 
 def disable():
     "disable global demand-loading of modules"
-    __builtins__["__import__"] = _origimport
+    __builtin__.__import__ = _origimport
 
