@@ -16,6 +16,14 @@
 import os
 import sys
 
+try:
+    import msvcrt
+    msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
+    msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
+    msvcrt.setmode(sys.stderr.fileno(), os.O_BINARY)
+except ImportError:
+    pass
+
 exitcode = 0
 out = sys.stdout
 
@@ -38,6 +46,9 @@ if url.startswith("file:"):
 elif url.startswith("remote:http"):
     os.environ["HG_URL"] = "remote:http"
 
+if "HG_PENDING" in os.environ:
+    os.environ["HG_PENDING"] = os.environ["HG_PENDING"] and "true"
+
 out.write("%s hook: " % name)
 for v in env:
     out.write("%s=%s " % (v, os.environ[v]))
