@@ -11,7 +11,22 @@ import difflib
 import errno
 import optparse
 import os
-import popen2
+try:
+    import subprocess
+    subprocess.Popen  # trigger ImportError early
+    closefds = os.name == 'posix'
+    def Popen4(cmd, bufsize=-1):
+        p = subprocess.Popen(cmd, shell=True, bufsize=bufsize,
+                             close_fds=closefds,
+                             stdin=subprocess.PIPE, stdout=subprocess.PIPE,
+                             stderr=subprocess.STDOUT)
+        p.fromchild = p.stdout
+        p.tochild = p.stdin
+        p.childerr = p.stderr
+        return p
+except ImportError:
+    subprocess = None
+    from popen2 import Popen4
 import shutil
 import signal
 import sys
@@ -63,14 +78,19 @@ parser.add_option("--tmpdir", type="string",
     help="run tests in the given temporary directory")
 parser.add_option("-v", "--verbose", action="store_true",
     help="output verbose messages")
+parser.add_option("-n", "--nodiff", action="store_true",
+    help="skip showing test changes")
 parser.add_option("--with-hg", type="string",
     help="test existing install at given location")
+parser.add_option("--pure", action="store_true",
+    help="use pure Python code instead of C extensions")
 
 for option, default in defaults.items():
     defaults[option] = int(os.environ.get(*default))
 parser.set_defaults(**defaults)
 (options, args) = parser.parse_args()
 verbose = options.verbose
+nodiff = options.nodiff
 coverage = options.cover or options.cover_stdlib or options.annotate
 python = sys.executable
 
@@ -78,8 +98,8 @@ if options.jobs < 1:
     print >> sys.stderr, 'ERROR: -j/--jobs must be positive'
     sys.exit(1)
 if options.interactive and options.jobs > 1:
-    print >> sys.stderr, 'ERROR: cannot mix -interactive and --jobs > 1'
-    sys.exit(1)
+    print '(--interactive overrides --jobs)'
+    options.jobs = 1
 
 def rename(src, dst):
     """Like os.rename(), trade atomicity and opened files friendliness
@@ -172,13 +192,14 @@ def install_hg():
     global python
     vlog("# Performing temporary installation of HG")
     installerrs = os.path.join("tests", "install.err")
+    pure = options.pure and "--pure" or ""
 
     # Run installer in hg root
     os.chdir(os.path.join(os.path.dirname(sys.argv[0]), '..'))
-    cmd = ('%s setup.py clean --all'
-           ' install --force --home="%s" --install-lib="%s"'
+    cmd = ('%s setup.py %s clean --all'
+           ' install --force --prefix="%s" --install-lib="%s"'
            ' --install-scripts="%s" >%s 2>&1'
-           % (sys.executable, INST, PYTHONDIR, BINDIR, installerrs))
+           % (sys.executable, pure, INST, PYTHONDIR, BINDIR, installerrs))
     vlog("# Running", cmd)
     if os.system(cmd) == 0:
         if not verbose:
@@ -202,6 +223,20 @@ def install_hg():
     os.environ["PYTHONPATH"] = pythonpath
 
     use_correct_python()
+    global hgpkg
+    hgpkg = _hgpath()
+
+    vlog("# Installing dummy diffstat")
+    f = open(os.path.join(BINDIR, 'diffstat'), 'w')
+    f.write('#!' + sys.executable + '\n'
+            'import sys\n'
+            'files = 0\n'
+            'for line in sys.stdin:\n'
+            '    if line.startswith("diff "):\n'
+            '        files += 1\n'
+            'sys.stdout.write("files patched: %d\\n" % files)\n')
+    f.close()
+    os.chmod(os.path.join(BINDIR, 'diffstat'), 0700)
 
     if coverage:
         vlog("# Installing coverage wrapper")
@@ -253,7 +288,7 @@ def run(cmd):
     """Run command in a sub-process, capturing the output (stdout and stderr).
     Return the exist code, and output."""
     # TODO: Use subprocess.Popen if we're running on Python 2.4
-    if os.name == 'nt':
+    if os.name == 'nt' or sys.platform.startswith('java'):
         tochild, fromchild = os.popen4(cmd)
         tochild.close()
         output = fromchild.read()
@@ -261,7 +296,7 @@ def run(cmd):
         if ret == None:
             ret = 0
     else:
-        proc = popen2.Popen4(cmd)
+        proc = Popen4(cmd)
         try:
             output = ''
             proc.tochild.close()
@@ -294,7 +329,8 @@ def run_one(test, skips, fails):
 
     def fail(msg):
         fails.append((test, msg))
-        print "\nERROR: %s %s" % (test, msg)
+        if not nodiff:
+            print "\nERROR: %s %s" % (test, msg)
         return None
 
     vlog("# Test", test)
@@ -344,7 +380,9 @@ def run_one(test, skips, fails):
         if os.name == 'nt':
             return skip("shell script")
         # do not try to run non-executable programs
-        if not os.access(testpath, os.X_OK):
+        if not os.path.exists(testpath):
+            return fail("does not exist")
+        elif not os.access(testpath, os.X_OK):
             return skip("not executable")
         cmd = '"%s"' % testpath
 
@@ -358,6 +396,8 @@ def run_one(test, skips, fails):
     if options.timeout > 0:
         signal.alarm(0)
 
+    mark = '.'
+
     skipped = (ret == SKIPPED_STATUS)
     # If reference output file exists, check test output against it
     if os.path.exists(ref):
@@ -367,22 +407,26 @@ def run_one(test, skips, fails):
     else:
         ref_out = []
     if skipped:
+        mark = 's'
         missing = extract_missing_features(out)
         if not missing:
             missing = ['irrelevant']
         skip(missing[-1])
     elif out != ref_out:
+        mark = '!'
         if ret:
             fail("output changed and returned error code %d" % ret)
         else:
             fail("output changed")
-        show_diff(ref_out, out)
+        if not nodiff:
+            show_diff(ref_out, out)
         ret = 1
     elif ret:
+        mark = '!'
         fail("returned error code %d" % ret)
 
     if not verbose:
-        sys.stdout.write(skipped and 's' or '.')
+        sys.stdout.write(mark)
         sys.stdout.flush()
 
     if ret != 0 and not skipped:
@@ -433,9 +477,11 @@ if not options.child:
 os.environ['LANG'] = os.environ['LC_ALL'] = 'C'
 os.environ['TZ'] = 'GMT'
 os.environ["EMAIL"] = "Foo Bar <foo.bar@example.com>"
+os.environ['CDPATH'] = ''
 
 TESTDIR = os.environ["TESTDIR"] = os.getcwd()
-HGTMP = os.environ['HGTMP'] = tempfile.mkdtemp('', 'hgtests.', options.tmpdir)
+HGTMP = os.environ['HGTMP'] = os.path.realpath(tempfile.mkdtemp('', 'hgtests.',
+                                               options.tmpdir))
 DAEMON_PIDS = None
 HGRCPATH = None
 
@@ -452,13 +498,25 @@ if options.with_hg:
     INST = options.with_hg
 else:
     INST = os.path.join(HGTMP, "install")
-BINDIR = os.path.join(INST, "bin")
+BINDIR = os.environ["BINDIR"] = os.path.join(INST, "bin")
 PYTHONDIR = os.path.join(INST, "lib", "python")
 COVERAGE_FILE = os.path.join(TESTDIR, ".coverage")
 
+def _hgpath():
+    cmd = '%s -c "import mercurial; print mercurial.__path__[0]"'
+    hgpath = os.popen(cmd % python)
+    path = hgpath.read().strip()
+    hgpath.close()
+    return path
+
+expecthg = os.path.join(HGTMP, 'install', 'lib', 'python', 'mercurial')
+hgpkg = None
+
 def run_children(tests):
     if not options.with_hg:
         install_hg()
+        if hgpkg != expecthg:
+            print '# Testing unexpected mercurial: %s' % hgpkg
 
     optcopy = dict(options.__dict__)
     optcopy['jobs'] = 1
@@ -512,6 +570,9 @@ def run_children(tests):
         print "Skipped %s: %s" % (s[0], s[1])
     for s in fails:
         print "Failed %s: %s" % (s[0], s[1])
+
+    if hgpkg != expecthg:
+        print '# Tested unexpected mercurial: %s' % hgpkg
     print "# Ran %d tests, %d skipped, %d failed." % (
         tested, skipped, failed)
     sys.exit(failures != 0)
@@ -525,6 +586,9 @@ def run_tests(tests):
         if not options.with_hg:
             install_hg()
 
+            if hgpkg != expecthg:
+                print '# Testing unexpected mercurial: %s' % hgpkg
+
         if options.timeout > 0:
             try:
                 signal.signal(signal.SIGALRM, alarmed)
@@ -585,6 +649,8 @@ def run_tests(tests):
                 print "Skipped %s: %s" % s
             for s in fails:
                 print "Failed %s: %s" % s
+            if hgpkg != expecthg:
+                print '# Tested unexpected mercurial: %s' % hgpkg
             print "# Ran %d tests, %d skipped, %d failed." % (
                 tested, skipped, failed)
 
