@@ -1,11 +1,10 @@
 import os
-from mercurial import hg, ui, commands
+from mercurial import hg, ui
 
 u = ui.ui()
 
 repo = hg.repository(u, 'test1', create=1)
 os.chdir('test1')
-repo = hg.repository(u, '.') # FIXME: can't lock repo without doing this
 
 # create 'foo' with fixed time stamp
 f = file('foo', 'w')
@@ -17,4 +16,4 @@ os.utime('foo', (1000, 1000))
 repo.add(['foo'])
 repo.commit(text='commit1', date="0 0")
 
-print "workingfilectx.date =", repo.workingctx().filectx('foo').date()
+print "workingfilectx.date =", repo[None]['foo'].date()
